/*Usage: sharpinteg 0hwp.fits 22.5hwp.fits 45hwp.fits 67.5hwp.fits -r rgmfile {-c -f #}*/
/* -c : Use Chauvenet's theorem for deglitching */
/* -f # : Manually set gain between h and v arrays */
/* version 1 edit - fixed array offset initialization, properly applied RGM, and fixed CRPIX1 and CRPIX2 discrepancy and set masked pixels to NAN*/
#define _GNU_SOURCE
#include <stdio.h>
#include <math.h>

#include "nrutil.h"
#include "/home/lkirby/cfitsio/fitsio.h"
#include "/home/lkirby/cfitsio/longnam.h"
#define  SCALE_NAS  4.57
#define SCALE_CASS 4.36
#define MAX_FILES 4
#define BLANK NAN
#define  ARRAY_ROWS  12
#define  ARRAY_COLUMNS  32
#define  SHARP_COLUMNS 12
#define  MAX_LEN  256
#define  MAX_O_TERMS  50
#define  BAD_MASK  7  
#define  PI  3.141592653589
#define  MAX_V  3.5
#define  MIN_V  -3.5
#define  MAX_TIME_JUMP  1.9
#define  CHOP_TOL0  4.0
#define  CHOP_SCALE  1.0
#define    Amid(x,lo,hi)  (((x)>=(lo))&&((x)<=(hi)))
double chauvenet(long N);
int sigma(double sumsq,double sum, int n, double *mean, double *var);
int read_float_column(fitsfile *fptr, char *name, long nrows, long *ncols,float *array);
int read_double_column(fitsfile *fptr, char *name, long nrows, long *ncols,double *array);
int read_logical_column(fitsfile *fptr, char *name, long nrows, long *ncols,char *array);
void printabort(char *message);
void printerror(int status);
void check_saturation(float *rawdata, int baddet[], long ndet, long nsamp,char *badsamp);
void check_timing(double *ut, double *det_time, long nsamp, char *badsamp,float sampdur);
void calc_tran(float skytran[], long nsamp, char *badsamp, float el[],float tauzen);
void calc_chop(float *chop, long nsamp, float chop_tol, float *chop_hi,float *chop_lo, float *chop_eff);
void chop_demod(float *chop, long *nsamp, float chop_hi, float chop_lo,float chop_tol, char *badsamp, double **rawdata, long ndet,float **el, float **ra, float **dec, float **azerr,float **elerr, float **azo, float **elo, float **rao,float **deco, float **pa, char **scanning, char **on_beam,char **off_beam);


int  IM_X=201, IM_Y=201;
float  IM_X0=101.0, IM_Y0=101.0;
float  ROTATOR_0 = 16.0;
float  MAGNIFICATION = 2.0;    /* map pixels/detector */
char  map_too_small = 'n';
float  A_CEN_X, A_CEN_Y;
float  SCALE = SCALE_CASS;     /* arcsec/detector */
char Usage[]="Usage: sharpinteg 0hwp.fits 22.5hwp.fits 45hwp.fits 67.5hwp.fits -r rgmfile {-c -f # -t # -a # #}\n   -c : Use Chauvenet's theorem for deglitching\n   -f # : Manually set gain between h and v arrays\n    -t  # : Manually set chop tolerance\n   -a # # :  X and Y offsets to align V array with H";

struct Header {
  char source[256];
  char platform[256];
  char date[256];
  float raep;
  float decep;
  float equinox;
  float rao0[MAX_FILES];
  float deco0[MAX_FILES];
  float azo0[MAX_FILES];
  float zao0[MAX_FILES];
  float fazo[MAX_FILES];
  float fzao[MAX_FILES];
  float rotator[MAX_FILES];
  float azo_chop[MAX_FILES];
  float rotator0[MAX_FILES];
  char hwp_ang[80];
};






int main(int Argc, char *Argv[])
{
  float year,raep,decep,*skytran[MAX_FILES],tauzen;
  float *bgnd_offset, *bgnd_rms, *mean_pa, *mean_el, *rawdata[MAX_FILES], *el[MAX_FILES], *ra[MAX_FILES], *dec[MAX_FILES], *azerr[MAX_FILES], *elerr[MAX_FILES], *azo[MAX_FILES], *elo[MAX_FILES], *rao[MAX_FILES], *deco[MAX_FILES], *pa[MAX_FILES], *nodstate[MAX_FILES], *chop[MAX_FILES], *source_state[MAX_FILES];
double *poldata[MAX_FILES];
  double *left[MAX_FILES],*leftsq[MAX_FILES],*right[MAX_FILES],*rightsq[MAX_FILES],*rightlow,*righthigh,*leftlow,*lefthigh,*raw[MAX_FILES];
  double A_CEN_X, A_CEN_Y,A_CEN_X_H,A_CEN_X_V,sampdur;
float *offset[MAX_FILES*MAX_O_TERMS];
  float  chop_throw = -1.0, chop_tol = CHOP_TOL0,chop_hi, chop_lo, chop_eff,val;
  float f[ARRAY_ROWS*SHARP_COLUMNS],fmed,ftemp[ARRAY_ROWS*SHARP_COLUMNS],fset,*xlist,*ylist,*sigx,*sigy,slope[MAX_FILES][SHARP_COLUMNS*ARRAY_ROWS],intercept,sigs,sigi,chi2,z,tempslope;
  double *polsig[MAX_FILES],*sig,*polerr[MAX_FILES],*sigerr,*demod[MAX_FILES],*demoderr[MAX_FILES],*sigsumerr,*u,*q,*uerr,*qerr,*rgm;
  double *ut, *det_time,leftmean,leftvar,rightmean,rightvar,rthres=3.0,lthres=3.0;
  long row,nrow,nsamp[MAX_FILES],ndet,width,ncol0,det,samp,*leftcount[MAX_FILES],*rightcount[MAX_FILES],npdet,nppdet,naxes[3],nelements,fpixel[3];
  char *scanning[MAX_FILES], *off_beam[MAX_FILES], *on_beam[MAX_FILES],*badsamp[MAX_FILES];
  char headline[512],platform[512],nasmyth_flag,source[512],*extname,*line,headline2[256];
  char *filename,*infiles[4],*rgmfile,deglitchflag=' ',fflag=' ';
  int  status = 0,extver,hdunum,hdutype,firstdatahdu,naxis,bitpix;   /* must be initialized to zero! */
  int colnum,typecode,term,baddet[ARRAY_ROWS*ARRAY_COLUMNS+1],nfiles;
  int i,j,ndetcol,ndetrow,arg,x,y,*bad,*badsharp,xoff=0,yoff=0;
  u_int err;
  FILE *rgmf,*fout;;
  fitsfile  *infptr,*outfptr;
  struct Header head;

  float testarray[144];
  if(Argc < 7) printabort(Usage);
  fprintf(stderr,"This is sharp integ version 1 last edited Mar 30, 2006\n");
  for(i=0;i<4;i++)
    {
      infiles[i]=Argv[i+1];
    }
  arg=5;
  if(strncmp(Argv[arg],"-r",2)==0)
    {
      arg++;
      arg++;
      rgmf=fopen(rgmfile=Argv[arg-1],"r");
    }
  else printabort(Usage);
  while(arg<Argc)
    {
     if(strncmp(Argv[arg],"-c",2)==0)
       {
	 arg++;
	 deglitchflag='c';
       }
     else if(strncmp(Argv[arg],"-f",2)==0)
	 {
	   arg++;
	   fflag='y';
	   sscanf(Argv[arg],"%f\n",&fset);
	   arg++;
	 }
     else if(strncmp(Argv[arg],"-t",2)==0)
	 {
	   arg++;
	   sscanf(Argv[arg],"%f\n",&chop_tol);
	   arg++;
	 }
     else if(strncmp(Argv[arg],"-a",2)==0)
       {
	 arg++;
	 sscanf(Argv[arg],"%d\n",&xoff);
	 arg++;
	 sscanf(Argv[arg],"%d\n",&yoff);
	 arg++;
       }
     else 
       {
	 printabort(Usage);
       }
    }
  for (i=0;i<4;i++)
    {
      filename=infiles[i];
      fprintf(stderr, "Reading raw data...%s\n",filename);
      fits_open_file(&infptr, filename, READONLY, &status);
      /* Read HDU 0 */
      if (fits_read_key(infptr, TSTRING, "OBJECT", head.source, headline, &status))
	{
	  printerror(status);
	}
      nasmyth_flag = 'n';
      if (fits_read_key(infptr, TSTRING, "PLATFORM", head.platform, headline, &status))
	{
	  printerror(status);
	}
      else 
	{
	  fprintf(stderr,"%s",head.platform);
	  if (strncmp(platform, "NA", 2) == 0) 
	    {
	      nasmyth_flag = 'y';
	      fprintf(stderr, "  Nasmyth platform detected\n");
	      SCALE = SCALE_NAS;     /* only one SCALE allowed per reduction */
	    }
	}
      if (fits_read_key(infptr, TSTRING, "DATE-OBS", head.date, headline, &status))
	{
	  printerror(status);
	}
      if (sscanf(head.date, "%d", &year) != 1)
	{
	  fprintf(stderr,"Could not read date from FITS file");
	}  
      if (fits_read_key(infptr, TFLOAT, "RAEP", &head.raep, headline, &status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "DECEP", &head.decep, headline, &status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "EQUINOX", &head.equinox, headline, &status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "RAO", &head.rao0[nfiles], headline,&status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "DECO", &head.deco0[nfiles], headline,&status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "AZO", &head.azo0[nfiles], headline,&status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "ZAO", &head.zao0[nfiles], headline,&status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "FAZO", &head.fazo[nfiles], headline,&status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "FZAO", &head.fzao[nfiles], headline,&status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "ROTATOR", &head.rotator[nfiles], headline,&status))
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TFLOAT, "AZO_CHOP", &head.azo_chop[nfiles],headline, &status)) 
	{
	  printerror(status);
	  head.azo_chop[nfiles] = 0.0;
	}
      if (nasmyth_flag == 'y') head.rotator[nfiles] = 0.0;
      if (fits_read_key(infptr, TFLOAT, "ROTZERO", &head.rotator0, headline,&status)) 
	{   /* Seems to handle string keyword OK */
	  status = 0;     /* November 2002 data doesn't have ROTZERO in header */
	}
      else 
	{
	  ROTATOR_0 = head.rotator0[nfiles];
	}
      if ((head.rotator[nfiles] < -360.0) || (head.rotator[nfiles] > 360.0)) 
	{
	  head.rotator[nfiles] = ROTATOR_0;
	  fprintf(stderr, "  WARNING:  Bad rotator value, setting to ");
	  system("date >> sharpsolve_errlog.txt");
	  system("echo \"WARNING:  Bad rotator value\" >> sharpsolve_errlog.txt");
	}
      head.rotator[nfiles] = ROTATOR_0-head.rotator[nfiles];
      fprintf(stderr, "  Cal. inst. rot. = %6.1f (file %d)\n", head.rotator[nfiles],nfiles+1);
      /* Get HWP angle from Hardware HDU */
      extver = 0;
      extname = "SHARC2 Hardware";
      if (fits_movnam_hdu(infptr, ANY_HDU, extname, extver, &status))
	{
	  printerror(status);
	}
      fits_get_hdu_num(infptr, &hdunum);
      if (hdunum != 2) 
	{
	  sprintf(line,"Warning: HDUNUM of Hardware extension in file %s is not 2.",filename);
	  printabort(line);
	}
      if (fits_read_card(infptr, "HWP_ANG", head.hwp_ang, &status)) 
	{
	  fprintf(stderr,"No HWP IN HEADER");
	  printerror(status);
	}
      hdunum = 5;
      if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) != 0)
	{
	  printerror(status);
	}
      if (fits_read_key(infptr, TSTRING, "EXTNAME", headline, headline2,&status))
	{
	  printerror(status);
	}
      firstdatahdu = 5;
      if (strncmp(headline, "SHARC2 Data", 11) != 0)
	{
	  firstdatahdu = 6;     /* hdunum = 5 is presumably CSO DSOS */
	}
      /* count total number of frames */
      hdunum = firstdatahdu;
      row = 0;
      while (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) == 0) 
	{
	  if (hdutype != BINARY_TBL)
	    {
	      printabort("Data HDU of wrong type");
	    }
	  if (fits_get_num_rows(infptr, &nrow, &status))
	    {
	      printerror(status);
	    }
	  row += nrow;
	  hdunum++;
	}
      nsamp[nfiles] = row;
      fprintf(stderr,"total number of rows in file is %d\n",nsamp[nfiles]);
      status = 0;
      /* allocate memory */
      hdunum = firstdatahdu;
      if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status))
	{
	  printerror(status);
	}
      if (hdutype != BINARY_TBL)
	{
	  printabort("Data HDU of wrong type");
	}
      if (fits_get_colnum(infptr, CASEINSEN, "Pixel Readout", &colnum, &status))
	{
	  printerror(status);
	}
      if (fits_get_coltype(infptr, colnum, &typecode, &ndet, &width, &status))
	{
	  printerror(status);
	}
      if (ndet != (ARRAY_ROWS*ARRAY_COLUMNS))
	{
	  printabort("Unexpected number of detectors");
	}

      rgm = (double *) malloc(ndet*sizeof(double));
      bad=(int *) malloc(ndet*sizeof(int));

      
      for(j=0;j<ndet;j++)
	{
	 
	  fscanf(rgmf,"%d %d %lf %d %*lf\n",&x,&y,&(rgm[j]),&(bad[j]));
	  if((x-1)*32+y != j+1) 
	    {
	      fprintf(stderr,"flaw in reading rgm file %d %d %d %d\n",x,y,(x-1)*32+y,j+1);
	    }
	}
      fclose(rgmf);
      rgmf=fopen(rgmfile,"r");
      if(ndet==384) npdet=288;
      else fprintf(stderr,"unknown number of polarimeter pixels\n");
      bgnd_offset = (float *) malloc(nfiles*sizeof(float));
      bgnd_rms = (float *) malloc(nfiles*sizeof(float));
      mean_pa = (float *) malloc(nfiles*sizeof(float));
      mean_el = (float *) malloc(nfiles*sizeof(float));
      rawdata[nfiles] = (float *) malloc(nsamp[nfiles]*ndet*sizeof(float));
      el[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      ra[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      dec[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      azerr[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      elerr[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      azo[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      elo[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      rao[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      deco[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      pa[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      scanning[nfiles] = (char *) malloc(nsamp[nfiles]*sizeof(char));
      on_beam[nfiles] = (char *) malloc(nsamp[nfiles]*sizeof(char));
      off_beam[nfiles] = (char *) malloc(nsamp[nfiles]*sizeof(char));
      nodstate[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      chop[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      source_state[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      ut = (double *) malloc(nsamp[nfiles]*sizeof(double));
      det_time = (double *) malloc(nsamp[nfiles]*sizeof(double));
      if (fits_read_key(infptr, TFLOAT, "CRPIX3", &A_CEN_Y, headline,&status))
	{
	  /*This is only for JSharc data, I believe*/
	  fprintf(stderr,"here %lf\n", A_CEN_Y);
	  status = 0;
	  A_CEN_X = (ARRAY_COLUMNS - 1.0)/2.0;
	  A_CEN_Y = (ARRAY_ROWS - 1.0)/2.0;
	}
      else 
	{
	  A_CEN_Y -= 1.0;
	  if (fits_read_key(infptr, TFLOAT, "CRPIX2", &A_CEN_X, headline,
			    &status))
	    printabort("Trouble reading CRPIX2");
	  A_CEN_X -= 1.0;
	}
      A_CEN_X_H = (SHARP_COLUMNS - 1.0)/2.0;
      A_CEN_X_V = (ARRAY_COLUMNS - 1.0) - (SHARP_COLUMNS - 1.0)/2.0;
      fprintf(stderr, "  Reference X pixels:  %.1f, %.1f\n", A_CEN_X_V+1,
              A_CEN_X_H+1);

      /*read data*/ 
      hdunum = firstdatahdu;
      row = 0;
      while (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) == 0) 
	{
	  if (hdutype != BINARY_TBL) printabort("Data HDU of wrong type");
	  if (fits_get_num_rows(infptr, &nrow, &status)) printerror(status);
	  if (read_float_column(infptr, "Pixel Readout", nrow, &ncol0,&(rawdata[nfiles][row*ndet]))) 
	    printerror(status);
	  if (read_float_column(infptr, "EL", nrow, &ncol0, &(el[nfiles][row]))) 
	    printerror(status);
	  if (read_float_column(infptr, "RA", nrow, &ncol0, &(ra[nfiles][row]))) 
	    printerror(status);
	  if (read_float_column(infptr, "DEC", nrow, &ncol0,&(dec[nfiles][row]))) 
	    printerror(status);
	  if (read_float_column(infptr, "AZ_ERROR", nrow, &ncol0,&(azerr[nfiles][row]))) 
	    printerror(status);
	  if (read_float_column(infptr, "EL_ERROR", nrow, &ncol0,&(elerr[nfiles][row]))) 
	    printerror(status);
	  if (read_float_column(infptr, "AZO", nrow, &ncol0,&(azo[nfiles][row])))
	    if (read_float_column(infptr, "AZ_OFFSET", nrow, &ncol0,&(azo[nfiles][row])))
	      printerror(status);
	  if (read_float_column(infptr, "ELO", nrow, &ncol0,&(elo[nfiles][row])))
	    if (read_float_column(infptr, "EL_OFFSET", nrow, &ncol0,&(elo[nfiles][row])))
	      printerror(status);
	  if (read_float_column(infptr, "RAO", nrow, &ncol0,&(rao[nfiles][row])))
	    if (read_float_column(infptr, "RA_OFFSET", nrow, &ncol0,&(rao[nfiles][row])))
	      printerror(status);
	  if (read_float_column(infptr, "DECO", nrow, &ncol0,&(deco[nfiles][row])))
	    if (read_float_column(infptr, "DEC_OFFSET", nrow, &ncol0,&(deco[nfiles][row])))
	      printerror(status);
	  if (read_float_column(infptr, "Parallactic Angle", nrow, &ncol0,&(pa[nfiles][row])))
	    printerror(status);
	  read_logical_column(infptr, "Scanning", nrow, &ncol0,&(scanning[nfiles][row]));
	  if (read_logical_column(infptr, "On Beam", nrow, &ncol0,&(on_beam[nfiles][row])))
	    {
	      status = 0;
	      /* WORK HERE */
	    }
	  if (read_logical_column(infptr, "Off Beam", nrow, &ncol0,&(off_beam[nfiles][row])))
	    {
	      status = 0;
	      /* WORK HERE */
	    }
	  if (read_float_column(infptr, "CHOP_OFFSET", nrow, &ncol0,&(chop[nfiles][row])))
	    printerror(status);
	  if (read_double_column(infptr, "UT", nrow, &ncol0, &(ut[row])))
	    printerror(status);
	  if (read_double_column(infptr, "Detector Time", nrow, &ncol0,&(det_time[row]))) 
	    {
	      status = 0;
	      det_time[0] = -1.0e30;     /* flag */
	    }
	  row += nrow;
	  hdunum++;
	}
      status = 0;
      if (fits_read_key(infptr, TFLOAT, "CDELT1", &sampdur, headline, &status))
	printerror(status);
      sampdur /= 1000.0;
      fprintf(stderr, "finished source = %s\n",head.source);
      /* Go to pixels HDU; SHOULD SEARCH FOR HDU TYPE IN FUTURE */
      hdunum = 4;
      if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status));
      printerror(status);
      if (hdutype != BINARY_TBL)
        printabort("Data HDU of wrong type");
      if (fits_get_num_rows(infptr, &nrow, &status))
        printerror(status);
      offset[nfiles*MAX_O_TERMS] = (float *) malloc(ndet*sizeof(float));
      if (read_float_column(infptr, "Pixel Offsets", nrow, &ncol0,offset[nfiles*MAX_O_TERMS]))
        printerror(status);
      if ((nrow*ncol0) != ndet)
        printabort("Unexpected number of detectors");
      for (term = 1; term < MAX_O_TERMS; term++) {
        offset[nfiles*MAX_O_TERMS+term] = (float *) malloc(ndet * sizeof(float));
        for (det = 1; det <= ndet; det++) 
	  {
	    offset[nfiles*MAX_O_TERMS+term][det-1] = 0.0;
	  }
      }

      if (fits_close_file(infptr, &status))
        printerror(status);
      
      badsamp[nfiles] = (char *) malloc(nsamp[nfiles] * sizeof(char));
      skytran[nfiles] = (float *) malloc(nsamp[nfiles] * sizeof(float));
      calc_tran(skytran[nfiles], nsamp[nfiles],badsamp[nfiles], el[nfiles],tauzen);
      for (samp=0;samp<nsamp[nfiles];samp++) badsamp[nfiles][samp]=0;
      /*check a few things (stolen from sharpsolve may not really need)*/
      check_saturation(rawdata[nfiles], baddet, ndet, nsamp[nfiles],badsamp[nfiles]);
      check_timing(ut, det_time, nsamp[nfiles], badsamp[nfiles], sampdur);
      
      free (ut);
      /*demodulate the chop*/
      calc_chop(chop[i], nsamp[i], chop_tol, &chop_hi, &chop_lo,&chop_eff);
      fprintf(stderr,"Measured chop high, low (eff.):  %.1f, %.1f  (%.f%%)\n",chop_hi, chop_lo, chop_eff*100.0);
      fprintf(stderr, "Measured throw = %.1f arcsec\n",(chop_hi-chop_lo)*CHOP_SCALE);
fprintf(stdout,"Measured chop high, low (eff.):  %.1f, %.1f  (%.f%%)\n",chop_hi, chop_lo, chop_eff*100.0);
      fprintf(stdout, "Measured throw = %.1f arcsec\n",(chop_hi-chop_lo)*CHOP_SCALE);
/*      if(on_beam[0][0]==on_beam[0][1]) fprintf(stderr,"should work");*/

      raw[nfiles]=(double *)malloc(nsamp[nfiles]*ndet*sizeof(double));
     
      for(samp=0;samp<nsamp[nfiles]*ndet;samp++){
	
	if(bad[samp%ndet] !=0)
	  {
	    raw[nfiles][samp]=0;
	    /*if((samp%ndet)%ARRAY_ROWS < 12) rawdata[nfiles][samp+20]=0;
	      else if((samp%ndet)%ARRAY_ROWS >=20) rawdata[nfiles][samp-20]=0;*/
	  }
	else raw[nfiles][samp]=(double)-1*rawdata[nfiles][samp]/rgm[samp%ndet];
	
      }
     
      chop_demod(chop[i], &(nsamp[i]), chop_hi, chop_lo, chop_tol,badsamp[i], &(raw[i]), ndet, &(el[i]),&(ra[i]), &(dec[i]), &(azerr[i]), &(elerr[i]),&(azo[i]), &(elo[i]), &(rao[i]), &(deco[i]),&(pa[i]), &(scanning[i]), &(on_beam[i]), &(off_beam[i]));


      
      /* for(samp=0;samp<nsamp[nfiles];samp++)
	{
	fprintf(stdout,"%d, %f\n",samp,rawdata[nfiles][samp*ndet+2]);
	}
      */
 
      xlist= (float *) malloc((nsamp[nfiles]+1)*sizeof(float));
      ylist= (float *) malloc((nsamp[nfiles]+1)*sizeof(float));
      sigx = (float *) malloc((nsamp[nfiles]+1)*sizeof(float));
      sigy = (float *) malloc((nsamp[nfiles]+1)*sizeof(float));
    
      /*f calculation
      fprintf(stderr,"%f\n",rawdata[nfiles][0]);

      fprintf(stderr,"%f\n",rawdata[nfiles][0]);*/

      fprintf(stderr,"nsamp is %d\n",nsamp[i]);

      for(ndetrow=0;ndetrow<12;ndetrow++)
	{
	  for(ndetcol=0;ndetcol<12;ndetcol++)
	    {
	      if(ndetcol+xoff<12 && ndetcol+xoff>=0 && ndetrow+yoff<12 && ndetrow+yoff>=0) {	 
		for (samp=0;samp<nsamp[nfiles];samp++)
		  {
		    det=ndetrow*32+ndetcol;
		   
		    
		    xlist[samp+1]=raw[nfiles][samp*ndet+ndetrow*32+ndetcol];
		    ylist[samp+1]=raw[nfiles][samp*ndet+(ndetrow+yoff)*32+ndetcol+xoff+20];
		    /*values don't matter*/
		 
		    /*  fprintf(stderr,"%f %f\n",xlist[samp+1],ylist[samp+1]);*/

		    sigx[samp+1]=1.0e5;
		    sigy[samp+1]=1.0e5;
		    /* if (ndetcol==2) fprintf(stdout,"%d %f %f\n",ndetrow*32+ndetcol,xlist[samp],ylist[samp]);*/
		  }
	      
		slope[i][ndetrow*12+ndetcol]=0;
		if(bad[ndetrow*32+ndetcol]==0 && bad[(ndetrow+yoff)*32+ndetcol+20+xoff]==0) fitexy(xlist,ylist,nsamp[i],sigx,sigy,&intercept,&slope[i][ndetrow*12+ndetcol],&sigi,&sigs,&chi2,&z);

	      }
	    }
	}


      
      poldata[nfiles] = (double *) malloc(nsamp[nfiles]*288*sizeof(double));
      
      nfiles++;
    }
  badsharp=(int *) malloc(npdet*sizeof(int));
  for(det=0;det<npdet/2;det++)
    {
      for(i=0;i<3;i++)
	{
	  for(j=i+1;j<4;j++)
	    {
	      if(slope[j][det]<slope[i][det])
		{
		  tempslope=slope[i][det];
		  slope[i][det]=slope[j][det];
		  slope[j][det]=tempslope;
		}
	    }
	}
    }     
  if(strncmp(&fflag,"y",1)!=0) 
    {
      for(det=0;det<npdet/2;det++)
	{
	  f[det]=(slope[1][det]+slope[2][det])/2;
	  fprintf(stdout,"%d %f\n",det,f[det]);
	}
    }
  else 
    {
      for(det=0;det<npdet/2;det++)
	{
	  f[det]=fset;
	}
    }

  for(i=0;i<npdet/2-1;i++)
    {
      for(j=i+1;j<npdet/2;j++)
	{
	  if(f[j]<f[i])
	    {
	      tempslope=f[i];
	      f[i]=f[j];
	      f[j]=tempslope;
	    }
	}
    }

  i=0;
  while(f[i]<=0) i++;
  fout=fopen("fmedians","a");

for(ndetrow=0;ndetrow<12;ndetrow++)
    {
      for(ndetcol=0;ndetcol<12;ndetcol++)
	{
	  badsharp[12*ndetrow+ndetcol]=bad[ndetrow*32+ndetcol];
	  badsharp[12*ndetrow+ndetcol+144]=bad[ndetrow*32+ndetcol+20];
	}
    }


  for(i=0;i<4;i++)
    {
      for(samp=0;samp<nsamp[i];samp++)
	{
	  for(ndetrow=0;ndetrow<12;ndetrow++)
	    {
	      for(ndetcol=0;ndetcol<12;ndetcol++)
		{
		  if(ndetcol+xoff<12 && ndetcol+xoff>=0 && ndetrow+yoff<12 && ndetrow+yoff>=0)
		    {
		      poldata[i][samp*288+12*ndetrow+ndetcol]=raw[i][samp*ndet+ndetrow*32+ndetcol]-raw[i][samp*ndet+(ndetrow+yoff)*32+ndetcol+xoff+20]/f[12*ndetrow+ndetcol];

		      poldata[i][samp*288+12*ndetrow+ndetcol+144]=raw[i][samp*ndet+ndetrow*32+ndetcol]+raw[i][samp*ndet+(ndetrow+yoff)*32+ndetcol+xoff+20]/f[12*ndetrow+ndetcol];
		    }
		  else
		    {
		      badsharp[12*ndetrow+ndetcol]=10;
		      poldata[i][samp*288+12*ndetrow+ndetcol]=0;
		      poldata[i][samp*288+12*ndetrow+ndetcol+144]=0;
		    }
		}
	    }
	}
    }
  
  nppdet=npdet/2;
  for(i=0;i<4;i++)
    {

      left[i] = (double *)malloc(npdet*sizeof(double));
      right[i] = (double *)malloc(npdet*sizeof(double)); 
      leftsq[i] = (double *)malloc(npdet*sizeof(double));
      rightsq[i] = (double *)malloc(npdet*sizeof(double));
      polsig[i]=(double *)malloc(nppdet*sizeof(double));
      sig=(double *)malloc(nppdet*sizeof(double));
      polerr[i]=(double *)malloc(nppdet*sizeof(double));
      sigerr=(double *)malloc(nppdet*sizeof(double));
      demod[i]=(double *)malloc(npdet*sizeof(double));
      demoderr[i]=(double *)malloc(npdet*sizeof(double));
      sigsumerr=(double *)malloc(npdet*sizeof(double));
      q=(double *)malloc(npdet/2*sizeof(double));
      u=(double *)malloc(npdet/2*sizeof(double));
      qerr=(double *)malloc(npdet/2*sizeof(double));
      uerr=(double *)malloc(npdet/2*sizeof(double));
      leftcount[i]=(long *)malloc(npdet*sizeof(long));
      rightcount[i]=(long *)malloc(npdet*sizeof(long));
      leftlow=(double *)malloc(npdet*sizeof(double));
      lefthigh=(double *)malloc(npdet*sizeof(double));
      rightlow=(double *)malloc(npdet*sizeof(double));
      righthigh=(double *)malloc(npdet*sizeof(double));

      for (det=0;det<npdet;det++)
	{
	  left[i][det]=0;
	  leftsq[i][det]=0;
	  leftcount[i][det]=0;
	  leftlow[det]=0;
	  lefthigh[det]=0;
	  right[i][det]=0;
	  rightsq[i][det]=0;
	  righthigh[det]=0;
	  rightlow[det]=0;
	  rightcount[i][det]=0;
	  demod[i][det]=0;
	  demoderr[i][det]=0;
	}

      /*add left and right beams*/
      for(samp=0;samp<nsamp[i];samp++)
	{
	  if(on_beam[i][samp]) 
	    {
	      for(det=0;det<npdet;det++)
		{
		  left[i][det]+=poldata[i][samp*npdet+det];
		  leftsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		  leftcount[i][det]+=1;
		}
	    }
	  else if(off_beam[i][samp])
	    {
	      for(det=0;det<npdet;det++)
		{
		  right[i][det]+=poldata[i][samp*npdet+det];
		  rightsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		  rightcount[i][det]+=1;
		}
	    }
	}
      if(deglitchflag=='c')
	{
	  rthres=chauvenet(rightcount[i][0]);
	  lthres=chauvenet(leftcount[i][0]);
	}
      fprintf(stderr,"thresholds are %lf %lf\n",rthres,lthres);
      /*process left and right beams to gather statistics for deglitching*/
      for (det=0;det<npdet;det++)
	{
	  err+=sigma(leftsq[i][det],left[i][det],leftcount[i][det],&leftmean,&leftvar);
	  leftlow[det]=leftmean-lthres*sqrt(leftvar);
	  lefthigh[det]=leftmean+lthres*sqrt(leftvar);
	  
	  err+=sigma(rightsq[i][det],right[i][det],rightcount[i][det],&rightmean,&rightvar);
	  rightlow[det]=rightmean-rthres*sqrt(rightvar);
	  righthigh[det]=rightmean+rthres*sqrt(rightvar);
	  
	}
      for(det=0;det<npdet;det++)
	{      
	  left[i][det]=0;
	  leftsq[i][det]=0;
	  leftcount[i][det]=0;
	  right[i][det]=0;
	  rightsq[i][det]=0;
	  rightcount[i][det]=0;
	}
      /*combine deglitched nodpairs*/
      for(samp=0;samp<nsamp[i];samp++)
	{
	  if(on_beam[i][samp]) 
	    {
	      for(det=0;det<npdet;det++)
		{
		  if(Amid(poldata[i][samp*npdet+det],leftlow[det],lefthigh[det]))
		    {
		      /*if(det==43) fprintf(stdout,"left %lf %lf\n",poldata[i][samp*npdet+det],poldata[i][samp*npdet+det+144]);*/
		      left[i][det]+=poldata[i][samp*npdet+det];
		      leftsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		      leftcount[i][det]++;
		    }
		}
	    }
	  else if(off_beam[i][samp])
	    {
	      for(det=0;det<npdet;det++)
		{
		  if(Amid(poldata[i][samp*npdet+det],rightlow[det],righthigh[det]))
		    {
		      /*if(det==43) fprintf(stdout,"right %lf %lf\n",poldata[i][samp*npdet+det],poldata[i][samp*npdet+det+144]);*/
		      right[i][det]+=poldata[i][samp*npdet+det];
		      rightsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		      rightcount[i][det]++;
		    }
		}
	    }
	}
    }
  for(i=0;i<4;i++)
    {
      for(det=0;det<npdet/2;det++)
	{
	  sig[det]=0;
	  sigsumerr[det]=0;
	  polsig[i][det]=0;
	  polerr[i][det]=0;
	}
    }

  for(i=0;i<4;i++)
    {
      for(det=0;det<npdet;det++)
	{
	  err+=sigma(leftsq[i][det],left[i][det],leftcount[i][det],&leftmean,&leftvar);
	  err+=sigma(rightsq[i][det],right[i][det],rightcount[i][det],&rightmean,&rightvar);
	  /*if(det==43 || det==187) fprintf(stdout,"%d %lf %lf\n",det,leftmean,rightmean);*/
	  demod[i][det]=leftmean-rightmean;
	  demoderr[i][det]=sqrt(leftvar/leftcount[i][det]+rightvar/rightcount[i][det]);
	}
      for(det=0;det<npdet/2;det++)
	{
	  /*if(det==43) fprintf(stdout,"%lf %lf\n",demod[i][det],demod[i][det+npdet/2]);*/
	  polsig[i][det]=demod[i][det];
	  polerr[i][det]=demoderr[i][det];
	  sig[det]+=demod[i][det+npdet/2]/4;
	  sigsumerr[det]+=demod[i][det+npdet/2]*demod[i][det+npdet/2];
	  
	}
    }

  for(det=0;det<npdet/2;det++)
    {  
      sigerr[det]=sqrt(fabs(sigsumerr[det]-sig[det]*sig[det]*4)/3);
      /*if (det==77)fprintf(stderr,"sig and sigsum are %f %f %f\n",sig[det],sigsumerr[det],sigerr[det]);*/
      q[det]=0.5*(polsig[0][det]-polsig[2][det]);
      u[det]=0.5*(polsig[1][det]-polsig[3][det]);
      qerr[det]=0.5*sqrt((polerr[0][det]*polerr[0][det]+polerr[2][det]*polerr[2][det]));
      uerr[det]=0.5*sqrt((polerr[1][det]*polerr[1][det]+polerr[3][det]*polerr[3][det]));
    }
  for(det=0;det<npdet/2;det++){
     /*  fprintf(stderr,"%d %lf %lf %lf %lf\n",det,sig[det],q[det],u[det],sqrt((u[det]*u[det]+q[det]*q[det])/sig[det]/sig[det]));	 */
	if(badsharp[det] !=0 || badsharp[det+npdet/2] !=0)
	  {
	    sig[det]=NAN;
	    sigerr[det]=NAN;
	    u[det]=NAN;
	    uerr[det]=NAN;
	    q[det]=NAN;
	    qerr[det]=NAN;
	    badsharp[det]+=badsharp[det+npdet/2];
	  }
      }

  system("rm -f solve.fits");
  if (fits_create_file(&outfptr, "solve.fits", &status))
    printerror(status);
  bitpix = DOUBLE_IMG;
  naxis = 2;
  naxes[0] = 12;
  naxes[1] = 12;
  nelements = 144;
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  fpixel[0] = 1;
  fpixel[1] = 1;
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, sig, &status))
    printerror(status); 
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "SIGNAL","", &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "SOURCE", &(head.source),"",&status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "DATE", &(head.date),"",&status))
    printerror(status);
  if (fits_write_key(outfptr, TFLOAT, "EQUINOX", &(head.equinox), "", &status))
    printerror(status);
  val=el[2][0];
 if (fits_write_key(outfptr,TFLOAT,"TEL_ELEV",&val,"",&status))
    printerror(status);
 val=pa[2][0];
 if (fits_write_key(outfptr,TFLOAT,"PARA_ANG",&val,"",&status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "CTYPE1", "ARRAY-X", "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRVAL1", &(val), "", &status))
    printerror(status);
  /*  val = SCALE/-3600.0;*/
  val=1;
  if (fits_write_key(outfptr, TFLOAT, "CDELT1", &val, "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRPIX1", &val, "", &status))
    printerror(status);
  val=0.0;
  if (fits_write_key(outfptr, TFLOAT, "CROTA1", &val, "", &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "CTYPE2", "ARRAY-Y", "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRVAL2", &(val), "", &status))
    printerror(status);
  /*val = SCALE/3600.0;*/
  val=1;
  if (fits_write_key(outfptr, TFLOAT, "CDELT2", &val, "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRPIX2", &val, "", &status))
    printerror(status);
  val=0.0;
  if (fits_write_key(outfptr, TFLOAT, "CROTA2", &val, "", &status))
    printerror(status);
  val=head.raep;
  if (fits_write_key(outfptr, TFLOAT, "RA", &val, "", &status))
    printerror(status);
  val=head.decep;
  if (fits_write_key(outfptr, TFLOAT, "DEC", &val, "", &status))
    printerror(status);
  val=head.rao0[2];
  if (fits_write_key(outfptr, TFLOAT, "RAO", &val, "", &status))
    printerror(status);
  val=head.deco0[2];
  if (fits_write_key(outfptr, TFLOAT, "DECO", &val, "", &status))
    printerror(status);
  val=head.azo0[2];
  if (fits_write_key(outfptr, TFLOAT, "AZO", &val, "", &status))
    printerror(status);
  val=head.zao0[2];
  if (fits_write_key(outfptr, TFLOAT, "ZAO", &val, "", &status))
    printerror(status);
  val=head.rotator[0];
  if (fits_write_key(outfptr,TFLOAT, "DEWERANG", &val, "", &status))
    printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "SIGNAL ERROR","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, sigerr, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "Q","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, q, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "Q ERROR","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, qerr, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "U","", &status))
    printerror(status);
  for (i=0;i<nppdet;i++) u[i]=-1*u[i];
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, u, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "U ERROR","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, uerr, &status))
        printerror(status);
  for (i=0;i<144;i++) testarray[i]=sqrt(u[i]*u[i]+q[i]*q[i])/sig[i];
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "PIXEL MASK","", &status))
    printerror(status);
  det=0;
  if (fits_write_key(outfptr, TINT, "VALUE FOR GOOD PIXEL", &det, "", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TINT, fpixel, nelements, badsharp, &status))
  printerror(status);
  if (fits_open_file(&infptr, filename, READONLY, &status))
    printerror(status);
  if (fits_copy_hdu(infptr, outfptr, 0, &status))
    printerror(status);
  if (fits_close_file(infptr, &status))
    printerror(status);
  



  if (fits_close_file(outfptr, &status))
    printerror(status);
  return 0;
}
/***********************/

int sigma(double sumsq,double sum, int n, double *mean, double *var)
{
  if (n<2)
    {
      *mean=sum;
      *var=0.0;
      return(1);
    }
  else
    {
      *mean=sum/n;
      *var=(sumsq-*mean * *mean * n)/(n-1);
      return(0);
    }
}

/***********************/

double chauvenet(long N)
{
  double low=0.0,new,high,value,reject,root2,precision;
  double TH_PREC=0.0004;

  reject=1.0/(2.0*N);
  root2=sqrt(2.0);
  precision=TH_PREC*root2;
  high=10.0/root2;

  while( (high-low)>precision)
    {
      new=(low+high)/2.0;
      value=erfc(new);
      if(value>reject)
	low=new;
      else
	high=new;
    }
  return((low+high)/root2);
}

/***********************/
int read_float_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                      float *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    float  nulval = FLOATNULLVALUE;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TFLOAT) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(float)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}
/*******************/

int read_double_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                       double *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    double  nulval = DOUBLENULLVALUE;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TDOUBLE) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(double)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}

/*******************/

int read_logical_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                        char *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    char  nulval = 0;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TLOGICAL) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(char)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}

/*******************/

void printabort(char *message)

{
    fprintf(stderr, "ERROR:  %s\n", message);
    system("date >> sharpsolve_errlog.txt");
    system("echo \"ABORTING:  unrecoverable error\" >> sharpsolve_errlog.txt");
    exit (1);
}

/*******************/

void printerror(int status)

{
    if (status)
    {
       fits_report_error(stderr, status);
       system("date >> sharpsolve_errlog.txt");
       system("echo \"ABORTING:  FITS error\" >> sharpsolve_errlog.txt");
       exit (status);
    }
    return;
}

/*******************/
void calc_tran(float skytran[], long nsamp, char *badsamp, float el[],
               float tauzen)

{
    long  samp;

    for (samp = 1; samp <= nsamp; samp++) {
      skytran[samp-1] = exp(-1.0*tauzen/sin(el[samp-1]*PI/180.0));
      if (el[samp-1] < 0.1) {
        fprintf(stderr, "  WARNING:  Bad elevation in frame %d -- flagging\n",
                samp);
        badsamp[samp-1] = 1;
        skytran[samp-1] = 1.0e-6;
        system("date >> sharpsolve_errlog.txt");
        system("echo \"WARNING:  Bad elevation\" >> sharpsolve_errlog.txt");
      }
    }
}

/*******************/

void check_saturation(float *rawdata, int baddet[], long ndet, long nsamp,
                      char *badsamp)

{
    long  det, samp;
    char  ok_flag[ARRAY_ROWS*ARRAY_COLUMNS];
    float  raw;

    for (det = 1; det <= ndet; det++)
      ok_flag[det-1] = 'y';
    for (det = 1; det <= ndet; det++) {
      for (samp = 1; samp <= nsamp; samp++) {
        if (badsamp[samp-1] == 0) {
          raw = rawdata[(samp-1)*ndet + (det-1)];
          if ((raw > MAX_V) || (raw < MIN_V))
            ok_flag[det-1] = 'n';
        }
      }
    }
    for (det = 1; det <= ndet; det++) {
      if ((ok_flag[det-1] != 'y') && ((baddet[det] & BAD_MASK) == 0)) {
        fprintf(stderr,
                "  WARNING:  Detector (%2d, %2d) near saturation\n",
                (det-1)/ARRAY_COLUMNS+1, (det-1)%ARRAY_COLUMNS+1);
        system("date >> sharpsolve_errlog.txt");
        system(
        "echo \"WARNING:  Detector near saturation\" >> sharpsolve_errlog.txt");
      }
    }
}

/*******************/

void check_timing(double *ut, double *det_time, long nsamp, char *badsamp,
                  float sampdur)

{
    long  samp, minsamp, maxsamp;
    double  dut, mindut, maxdut;

    mindut = 1.0e30;
    maxdut = -1.0e30;
    for (samp = 1; samp < nsamp; samp++) {
      if ((badsamp[samp-1] == 0) && (badsamp[samp] == 0)) {
        dut = ut[samp] - ut[samp-1];
        if (dut < mindut) {
          mindut = dut;
          minsamp = samp+1;
        }
        if (dut > maxdut) {
          maxdut = dut;
          maxsamp = samp+1;
        }
      }
    }
    if (mindut < 0.0) {
      fprintf(stderr, "  WARNING:  Time not monotonic for sample %ld\n",
              minsamp);
      system("date >> sharpsolve_errlog.txt");
      system("echo \"WARNING:  Time not monotonic\" >> sharpsolve_errlog.txt");
    }
    maxdut *= 3600.0;
    if (maxdut > (MAX_TIME_JUMP*sampdur)) {
      fprintf(stderr, "  WARNING:  Time jump (%.3lf sec) for sample %ld\n",
              maxdut, maxsamp);
      system("date >> sharpsolve_errlog.txt");
      system("echo \"WARNING:  Time jump\" >> sharpsolve_errlog.txt");
    }
    mindut = 1.0e30;
    maxdut = -1.0e30;
    if (det_time[0] > -1.0e29) {
      for (samp = 1; samp <= nsamp; samp++) {
        if ((badsamp[samp-1] == 0)) {
          dut = det_time[samp-1] - ut[samp-1]*3600.0;     /* seconds */
          if (dut < mindut) {
            mindut = dut;
            minsamp = samp+1;
          }
          if (dut > maxdut) {
            maxdut = dut;
            maxsamp = samp+1;
          }
        }
      }
      dut = (maxdut + mindut)/2.0;
      dut = dut - 86400.0*((int) (dut/86400.0 + 0.5));
      fprintf(stderr, "  Post April 2003 timing check:  %.3f %.3f %.3f %.3f\n",
              mindut, maxdut, dut, maxdut - mindut);
      if (fabs(dut) > (MAX_TIME_JUMP*sampdur)) {
        fprintf(stderr,
          "  WARNING:  Synchronization offset:  %.3f sec\n", dut);
        system("date >> sharpsolve_errlog.txt");
        system(
         "echo \"WARNING:  Synchronization offset\" >> sharpsolve_errlog.txt");
      }
      if ((maxdut - mindut) > (MAX_TIME_JUMP*sampdur)) {
        fprintf(stderr,
          "  WARNING:  Synchronization glitch:  %.3f sec\n", maxdut-mindut);
        system("date >> sharpsolve_errlog.txt");
        system(
         "echo \"WARNING:  Synchronization glitch\" >> sharpsolve_errlog.txt");
      }
    }
}

/*******************/

void calc_chop(float *chop, long nsamp, float chop_tol, float *chop_hi,
               float *chop_lo, float *chop_eff)

{
    float  himean, lomean, hisig, losig;
    float  himax, himin, lomax, lomin;
    float  mean, sum0, hi0, lo0;
    long  samp;
    int  i;

    /* Should I exclude bad samples? */
    mean = 0.0;
    sum0 = 0.0;
    for (samp = 1; samp <= nsamp; samp++) {
      mean += chop[samp-1];
      sum0 += 1.0;
    }
    mean /= sum0;
    himax = 1.0e30;
    himin = mean;
    lomin = -1.0e30;
    lomax = mean;
    for (i = 1; i <= 10; i++) {
      himean = 0.0;
      lomean = 0.0;
      hi0 = 0.0;
      lo0 = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
       if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax)) {
          himean += chop[samp-1];
          hi0 += 1.0;
        }
        if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax)) {
          lomean += chop[samp-1];
          lo0 += 1.0;
        }
      }
      himean /= hi0;
      lomean /= lo0;
      hisig = 0.0;
      losig = 0.0;
      hi0 = 0.0;
      lo0 = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
        if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax)) {
          hisig += (chop[samp-1] - himean)*(chop[samp-1] - himean);
          hi0 += 1.0;
        }
        if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax)) {
          losig += (chop[samp-1] - lomean)*(chop[samp-1] - lomean);
          lo0 += 1.0;
        }
      }
      hisig = sqrt(hisig/hi0);
      losig = sqrt(losig/lo0);
      if (hisig < (chop_tol / 2.0))
        hisig = chop_tol / 2.0;
      if (losig < (chop_tol / 2.0))
        losig = chop_tol / 2.0;
      himax = himean + hisig;
      himin = himean - hisig;
      lomax = lomean + losig;
      lomin = lomean - losig;
    }
    *chop_hi = himean;
    *chop_lo = lomean;
    *chop_eff = (hi0 + lo0)/sum0;
    /* assumes that sigmas eventually become smaller than chop_tol / 2.0 */
}

/*******************/

void chop_demod(float *chop, long *nsamp, float chop_hi, float chop_lo,
                float chop_tol, char *badsamp, double **rawdata, long ndet,
                float **el, float **ra, float **dec, float **azerr,
                float **elerr, float **azo, float **elo, float **rao,
                float **deco, float **pa, char **scanning, char **on_beam, char **off_beam)

{
    long  *choptag;
    float  *hiweight, *loweight;
    float  himax, himin, lomax, lomin;
    long  hicount, locount;
    int  old_chop_state, new_chop_state,beamcheck;
    long  samp, nchop, samp0, samp1;
    double  *c_rawdata;
    float  *c_el, *c_ra, *c_dec, *c_azerr, *c_elerr;
    float  *c_azo, *c_elo, *c_rao, *c_deco, *c_pa; 
    char *c_off,*c_on;
    char  *c_scanning;
    long  det;

/*     for (samp=1; samp <= *nsamp; samp++) */
/*       fprintf(stderr,"%d, %lf\n", samp, (*nodstate)[samp-1]); */

    /* Doesn'T CORRECTLY USE SCANNING FLAG */
    /* SHOULD HANDLE PA DIFFERENTLY */

    choptag = (long *) malloc(*nsamp*sizeof(long));
    hiweight = (float *) malloc(*nsamp*sizeof(float));
    loweight = (float *) malloc(*nsamp*sizeof(float));

    himax = chop_hi + chop_tol/2.0;
    himin = chop_hi - chop_tol/2.0;
    lomax = chop_lo + chop_tol/2.0;
    lomin = chop_lo - chop_tol/2.0;
    fprintf(stderr,"chop tolerances:himax %f himin %f lomax %f lomin %f\n",himax,himin,lomax,lomin);
    hicount = 0;
    locount = 0;
    old_chop_state = 0;
    nchop = 0;
    samp0 = 1;
    for (samp = 1; samp <= *nsamp; samp++) {
      new_chop_state = 0;
      hiweight[samp-1] = 0.0;
      loweight[samp-1] = 0.0;
      if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax))
        new_chop_state = 1;
      if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax))
        new_chop_state = -1;
      if (new_chop_state != 0) {
	if((*on_beam)[samp-1]==(*on_beam)[samp0-1] && abs((*elerr)[samp-1]) < 5 && abs((*azerr)[samp-1]) < 5){
	  if ((new_chop_state*old_chop_state) == -1) {
	    if ((locount > (hicount/2)) && (hicount > (locount/2))) {
	      for (samp1 = samp0; samp1 < samp; samp1++) {
		hiweight[samp1-1] /= (float)hicount;
		loweight[samp1-1] /= (float)locount;
	      }
	      hicount = 0;
	      locount = 0;
	      nchop++;
	      samp0 = samp;
	    }
	    else {
	      for (samp1 = samp0; samp1 < samp; samp1++) {
		if (new_chop_state == 1)
		  hiweight[samp1-1] = 0.0;
		if (new_chop_state == -1)
		  loweight[samp1-1] = 0.0;
	      }
	      if (new_chop_state == 1)
		hicount = 0;
	      if (new_chop_state == -1)
		locount = 0;
	    }
	  }
      	}
	else {
	  for(samp1=samp0;samp1<samp;samp1++){
	    hiweight[samp1-1]=0;
	    loweight[samp1-1]=0;
	    }
	  hicount=0;
	  locount=0;
	  samp0=samp;
	  }
	if (badsamp[samp-1] == 0) {
	  if (new_chop_state == 1) {
	    hiweight[samp-1] = 1.0;
	    hicount++;
	  }
	  if (new_chop_state == -1) {
	    loweight[samp-1] = 1.0;
	    locount++;
	  }
	}
	old_chop_state = new_chop_state;
    }
      choptag[samp-1] = nchop;
      
    }
    if ((locount > (hicount/2)) && (hicount > (locount/2))) {
      for (samp1 = samp0; samp1 < samp; samp1++) {
        hiweight[samp1-1] /= (float)hicount;
        loweight[samp1-1] /= (float)locount;
      }
      hicount = 0;
      locount = 0;
      nchop++;
      samp0 = samp;
    }

    c_rawdata = (double *) malloc(nchop*ndet*sizeof(double));
    c_el = (float *) malloc(nchop*sizeof(float));
    c_ra = (float *) malloc(nchop*sizeof(float));
    c_dec = (float *) malloc(nchop*sizeof(float));
    c_azerr = (float *) malloc(nchop*sizeof(float));
    c_elerr = (float *) malloc(nchop*sizeof(float));
    c_azo = (float *) malloc(nchop*sizeof(float));
    c_elo = (float *) malloc(nchop*sizeof(float));
    c_rao = (float *) malloc(nchop*sizeof(float));
    c_deco = (float *) malloc(nchop*sizeof(float));
    c_pa = (float *) malloc(nchop*sizeof(float));
    c_scanning = (char *) malloc(nchop*sizeof(char));
    c_on = (char *) malloc(nchop*sizeof(float));
    c_off = (char *) malloc(nchop*sizeof(float));

    for (samp = 1; samp <= nchop; samp++) {
      for (det = 1; det <= ndet; det++) {
        c_rawdata[(samp-1)*ndet + (det-1)] = 0.0;
      }
      c_el[samp-1] = 0.0;
      c_ra[samp-1] = 0.0;
      c_dec[samp-1] = 0.0;
      c_azerr[samp-1] = 0.0;
      c_elerr[samp-1] = 0.0;
      c_azo[samp-1] = 0.0;
      c_elo[samp-1] = 0.0;
      c_rao[samp-1] = 0.0;
      c_deco[samp-1] = 0.0;
      c_pa[samp-1] = 0.0;
      c_scanning[samp-1] = 1;
      c_on[samp-1] = 0.0;
      c_off[samp-1]=0.0;
    }
   
    for (samp = 1; samp <= *nsamp; samp++) {
      if (choptag[samp-1] < nchop) {
        for (det = 1; det <= ndet; det++) {
          c_rawdata[choptag[samp-1]*ndet + (det-1)] +=
                                     (*rawdata)[(samp-1)*ndet + (det-1)] *
                                     (hiweight[samp-1]-loweight[samp-1]);
        }
        c_el[choptag[samp-1]] += (*el)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_ra[choptag[samp-1]] += (*ra)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_dec[choptag[samp-1]] += (*dec)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_azerr[choptag[samp-1]] += (*azerr)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_elerr[choptag[samp-1]] += (*elerr)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_azo[choptag[samp-1]] += (*azo)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_elo[choptag[samp-1]] += (*elo)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_rao[choptag[samp-1]] += (*rao)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_deco[choptag[samp-1]] += (*deco)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_pa[choptag[samp-1]] += (*pa)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
 
        c_on[choptag[samp-1]] = (*on_beam)[samp-1];
	c_off[choptag[samp-1]] = (*off_beam)[samp-1];
 	fprintf(stdout,"%d, %d, %d, %lf %lf, %lf\n",samp,choptag[samp-1],c_on[choptag[samp-1]],chop[samp-1],hiweight[samp-1], loweight[samp-1]);
      }
    }


    free(*rawdata);
    *rawdata = c_rawdata;
    free(*el);
    *el = c_el;
    free(*ra);
    *ra = c_ra;
    free(*dec);
    *dec = c_dec;
    free(*azerr);
    *azerr = c_azerr;
    free(*elerr);
    *elerr = c_elerr;
    free(*azo);
    *azo = c_azo;
    free(*elo);
    *elo = c_elo;
    free(*rao);
    *rao = c_rao;
    free(*deco);
    *deco = c_deco;
    free(*pa);
    *pa = c_pa;
    free(*scanning);
    *scanning = c_scanning;
    free(*on_beam);
    *on_beam = c_on;
    free(*off_beam);
    *off_beam=c_off;



    *nsamp = nchop;
    free(choptag);
    free(hiweight);
    free(loweight);

}

/*******************/
